open Type

(* Interface des arbres abstraits *)
module type Ast = sig
  type expression
  type instruction
  type fonction
  type programme
end

(* *************************************** *)
(* AST après la phase d'analyse syntaxique *)
(* *************************************** *)
module AstSyntax = struct
  (* Opérateurs unaires de Rat *)
  type unaire = Numerateur | Denominateur

  (* Opérateurs binaires de Rat *)
  type binaire = Fraction | Plus | Mult | Equ | Inf

  type assignment_target =
    | Ident of string
    | Deref of assignment_target
    | ArrayAccess of assignment_target * expression

  (* Expressions de Rat *)
  and expression =
    (* Appel de fonction représenté par le nom de la fonction et la liste des paramètres réels *)
    | AppelFonction of string * expression list
    (* Booléen *)
    | Booleen of bool
    (* Entier *)
    | Entier of int
    (* Opération unaire représentée par l'opérateur et l'opérande *)
    | Unaire of unaire * expression
    (* Opération binaire représentée par l'opérateur, l'opérande gauche et l'opérande droite *)
    | Binaire of binaire * expression * expression
    | AssignmentTarget of assignment_target
    (* null *)
    | Null
    (* &a *)
    | Address of string
    (* new T *)
    | NewPointer of typ
    (* new T[E] *)
    | NewArray of typ * expression
    (* {E1, E2, E2, ...} *)
    | ArrayLiteral of expression list

  (* Instructions de Rat *)
  type bloc = instruction list

  and instruction =
    (* Instruction portant une étiquette *)
    | Labelled of string * instruction 
    (* Déclaration de variable représentée par son type, son nom et l'expression d'initialisation *)
    | Declaration of typ * string * expression
    (* Affectation d'une variable représentée par son nom et la nouvelle valeur affectée *)
    | Affectation of assignment_target * expression
    (* Déclaration d'une constante représentée par son nom et sa valeur (entier) *)
    | Constante of string * int
    (* Affichage d'une expression *)
    | Affichage of expression
    (* Conditionnelle représentée par la condition, le bloc then et le bloc else *)
    | Conditionnelle of expression * bloc * bloc
    (*Boucle TantQue représentée par la conditin d'arrêt de la boucle et le bloc d'instructions *)
    | TantQue of expression * bloc
    (* return d'une fonction *)
    | Retour of expression
    (* boucle for représentée par la variable pour compter,
       la condition d'arrêt, l'incrément et le bloc d'instructions *)
    | For of string * expression * expression * string * expression * bloc
    (* instruction pour jump à une étiquette *)
    | Goto of string

  (* Structure des fonctions de Rat *)
  (* type de retour - nom - liste des paramètres (association type et nom) - corps de la fonction *)
  type fonction = Fonction of typ * string * (typ * string) list * bloc

  (* Structure d'un programme Rat *)
  (* liste de fonction - programme principal *)
  type programme = Programme of fonction list * bloc
end

(* ********************************************* *)
(* AST après la phase d'analyse des identifiants *)
(* ********************************************* *)
module AstTds = struct
  (* Expressions existantes dans notre langage *)

  type assignment_target =
    | Ident of Tds.info_ast
    | Deref of assignment_target
    | ArrayAccess of assignment_target * expression

  (* ~ expression de l'AST syntaxique où les noms des identifiants ont été
     remplacés par les informations associées aux identificateurs *)
  and expression =
    | AppelFonction of Tds.info_ast * expression list
      (* le nom de l'identifiant est remplacé par ses informations *)
    | Booleen of bool
    | Entier of int
    | Unaire of AstSyntax.unaire * expression
    | Binaire of AstSyntax.binaire * expression * expression
    | Null
    | AssignmentTarget of assignment_target
    | Address of Tds.info_ast
    | NewPointer of typ
    | NewArray of typ * expression
    | ArrayLiteral of expression list

  (* instructions existantes dans notre langage *)
  (* ~ instruction de l'AST syntaxique où les noms des identifiants ont été
     remplacés par les informations associées aux identificateurs
     + suppression de nœuds (const) *)
  type bloc = instruction list

  and instruction =
    | Declaration of typ * Tds.info_ast * expression
      (* le nom de l'identifiant est remplacé par ses informations *)
    | Affectation of assignment_target * expression
      (* le nom de l'identifiant est remplacé par ses informations *)
    | Affichage of expression
    | Conditionnelle of expression * bloc * bloc
    | TantQue of expression * bloc
    | Retour of expression * Tds.info_ast
      (* les informations sur la fonction à laquelle est associé le retour *)
    | Empty (* les nœuds ayant disparus: Const *)
    (* (initialisation, condition de continuation, incrément (instruction à chaque itération), bloc à exécuter) *)
    | For of instruction * expression * instruction * bloc
    | Goto of Tds.info_ast
    | Labelled of Tds.info_ast * instruction

  (* Structure des fonctions dans notre langage *)
  (* type de retour - informations associées à l'identificateur (dont son nom) - liste des paramètres (association type et information sur les paramètres) - corps de la fonction *)
  type fonction =
    | Fonction of typ * Tds.info_ast * (typ * Tds.info_ast) list * bloc

  (* Structure d'un programme dans notre langage *)
  type programme = Programme of fonction list * bloc
end

(* ******************************* *)
(* AST après la phase de typage *)
(* ******************************* *)
module AstType = struct
  (* Opérateurs unaires de Rat - résolution de la surcharge *)
  type unaire = Numerateur | Denominateur

  (* Opérateurs binaires existants dans Rat - résolution de la surcharge *)
  type binaire =
    | Fraction
    | PlusInt
    | PlusRat
    | MultInt
    | MultRat
    | EquInt
    | EquBool
    | Inf

  type assignment_target =
    | Ident of Tds.info_ast
    | Deref of typ * assignment_target
    | ArrayAccess of typ * assignment_target * expression

  (* Expressions existantes dans Rat *)
  (* = expression de AstTds *)
  and expression =
    | AppelFonction of Tds.info_ast * expression list
    | Booleen of bool
    | Entier of int
    | Unaire of unaire * expression
    | Binaire of binaire * expression * expression
    | AssignmentTarget of assignment_target
    | Address of Tds.info_ast
    | NewPointer of typ
    | NewArray of typ * expression
    | ArrayLiteral of typ * expression list
    | Null

  (* instructions existantes Rat *)
  (* = instruction de AstTds + informations associées aux identificateurs, mises à jour *)
  (* + résolution de la surcharge de l'affichage *)
  type bloc = instruction list

  and instruction =
    | Declaration of Tds.info_ast * expression
    | Affectation of assignment_target * expression
    | AffichageInt of expression
    | AffichageRat of expression
    | AffichageBool of expression
    | AffichageAddresse of assignment_target
    | Conditionnelle of expression * bloc * bloc
    | TantQue of expression * bloc
    | Retour of expression * Tds.info_ast
    | Empty (* les nœuds ayant disparus: Const *)
    | For of instruction * expression * instruction * bloc
    | Goto of Tds.info_ast
    | Labelled of Tds.info_ast * instruction

  (* informations associées à l'identificateur (dont son nom), liste des paramètres, corps *)
  type fonction = Fonction of Tds.info_ast * Tds.info_ast list * bloc

  (* Structure d'un programme dans notre langage *)
  type programme = Programme of fonction list * bloc
end

(* ******************************* *)
(* AST après la phase de placement *)
(* ******************************* *)
module AstPlacement = struct
  (* Expressions existantes dans notre langage *)
  (* = expression de AstType  *)
  type expression = AstType.expression
  type assignement_target = AstType.assignment_target

  (* instructions existantes dans notre langage *)
  type bloc = instruction list * int (* taille du bloc *)

  and instruction =
    | Labelled of Tds.info_ast * instruction
    | Declaration of Tds.info_ast * expression
    | Affectation of assignement_target * expression
    | AffichageInt of expression
    | AffichageRat of expression
    | AffichageBool of expression
    | AffichageAddresse of assignement_target
    | Conditionnelle of expression * bloc * bloc
    | TantQue of expression * bloc
    | Retour of
        expression * int * int (* taille du retour et taille des paramètres *)
    | Empty (* les nœuds ayant disparus: Const *)
    | For of instruction * expression * instruction * bloc
    | Goto of Tds.info_ast

  (* informations associées à l'identificateur (dont son nom), liste de paramètres, corps, expression de retour *)
  (* Plus besoin de la liste des paramètres mais on la garde pour les tests du placements mémoire *)
  type fonction = Fonction of Tds.info_ast * Tds.info_ast list * bloc

  (* Structure d'un programme dans notre langage *)
  type programme = Programme of fonction list * bloc
end
